from __future__ import annotations

import types
import typing
from typing import Any

from pybind11_stubgen.parser.errors import InvalidIdentifierError
from pybind11_stubgen.parser.interface import IParser
from pybind11_stubgen.structs import (
    Alias,
    Attribute,
    Class,
    Docstring,
    Field,
    Function,
    Identifier,
    Import,
    Method,
    Module,
    Property,
    QualifiedName,
)


class FilterTypingModuleAttributes(IParser):
    __sentinel = object()

    def handle_attribute(self, path: QualifiedName, attr: Any) -> Attribute | None:
        if getattr(typing, str(path[-1]), self.__sentinel) is attr:
            return None
        return super().handle_attribute(path, attr)


class FilterClassMembers(IParser):
    __attribute_blacklist: set[Identifier] = {
        *map(
            Identifier,
            (
                "__annotations__",
                "__builtins__",
                "__cached__",
                "__file__",
                "__loader__",
                "__name__",
                "__package__",
                "__path__",
                "__spec__",
            ),
        )
    }
    __class_member_blacklist: set[Identifier] = {
        *map(
            Identifier,
            (
                "__annotations__",
                "__class__",
                "__dict__",
                "__module__",
                "__qualname__",
                "__weakref__",
            ),
        )
    }

    __method_blacklist: set[Identifier] = {*map(Identifier, ("__dir__", "__sizeof__"))}

    def handle_class_member(
        self, path: QualifiedName, class_: type, member: Any
    ) -> Docstring | Alias | Class | list[Method] | Field | Property | None:
        name = path[-1]
        if name in self.__class_member_blacklist:
            return None
        if not hasattr(class_, "__dict__") or name not in class_.__dict__:
            # Skip members inherited from base classes
            return None
        return super().handle_class_member(path, class_, member)

    def handle_method(self, path: QualifiedName, value: Any) -> list[Method]:
        if path[-1] in self.__method_blacklist:
            return []
        return super().handle_method(path, value)

    def handle_attribute(self, path: QualifiedName, attr: Any) -> Attribute | None:
        if path[-1] in self.__attribute_blacklist:
            return None
        return super().handle_attribute(path, attr)


class FilterPybindInternals(IParser):
    __attribute_blacklist: set[Identifier] = {*map(Identifier, ("__entries",))}

    __class_blacklist: set[Identifier] = {*map(Identifier, ("pybind11_type",))}

    def handle_attribute(self, path: QualifiedName, value: Any) -> Attribute | None:
        if path[-1] in self.__attribute_blacklist:
            return None
        return super().handle_attribute(path, value)

    def handle_class_member(
        self, path: QualifiedName, class_: type, member: Any
    ) -> Docstring | Alias | Class | list[Method] | Field | Property | None:
        name = path[-1]
        if name in self.__class_blacklist:
            return None
        if name.startswith("__pybind11_module"):
            return None
        return super().handle_class_member(path, class_, member)


class FilterInvalidIdentifiers(IParser):
    def handle_module_member(
        self, path: QualifiedName, module: types.ModuleType, obj: Any
    ) -> (
        Docstring | Import | Alias | Class | list[Function] | Attribute | Module | None
    ):
        if not path[-1].isidentifier():
            self.report_error(InvalidIdentifierError(path[-1], path.parent))
            return None
        return super().handle_module_member(path, module, obj)

    def handle_class_member(
        self, path: QualifiedName, class_: type, obj: Any
    ) -> Docstring | Alias | Class | list[Method] | Field | Property | None:
        if not path[-1].isidentifier():
            self.report_error(InvalidIdentifierError(path[-1], path.parent))
            return None
        return super().handle_class_member(path, class_, obj)



from __future__ import annotations

import importlib
import logging
import re
from argparse import ArgumentParser
from pathlib import Path

from pybind11_stubgen.parser.interface import IParser
from pybind11_stubgen.parser.mixins.error_handlers import (
    IgnoreAllErrors,
    IgnoreInvalidExpressionErrors,
    IgnoreInvalidIdentifierErrors,
    IgnoreUnresolvedNameErrors,
    LogErrors,
    LoggerData,
    SuggestCxxSignatureFix,
    TerminateOnFatalErrors,
)
from pybind11_stubgen.parser.mixins.filter import (
    FilterClassMembers,
    FilterInvalidIdentifiers,
    FilterPybindInternals,
    FilterTypingModuleAttributes,
)
from pybind11_stubgen.parser.mixins.fix import (
    FixBuiltinTypes,
    FixCurrentModulePrefixInTypeNames,
    FixMissing__all__Attribute,
    FixMissing__future__AnnotationsImport,
    FixMissingEnumMembersAnnotation,
    FixMissingFixedSizeImport,
    FixMissingImports,
    FixMissingNoneHashFieldAnnotation,
    FixNumpyArrayDimAnnotation,
    FixNumpyArrayFlags,
    FixNumpyArrayRemoveParameters,
    FixNumpyDtype,
    FixPEP585CollectionNames,
    FixPybind11EnumStrDoc,
    FixRedundantBuiltinsAnnotation,
    FixRedundantMethodsFromBuiltinObject,
    FixTypingTypeNames,
    FixValueReprRandomAddress,
    OverridePrintSafeValues,
    RemoveSelfAnnotation,
    ReplaceReadWritePropertyWithField,
    RewritePybind11EnumValueRepr,
)
from pybind11_stubgen.parser.mixins.parse import (
    BaseParser,
    ExtractSignaturesFromPybind11Docstrings,
    ParserDispatchMixin,
)
from pybind11_stubgen.printer import Printer
from pybind11_stubgen.structs import QualifiedName
from pybind11_stubgen.writer import Writer


def arg_parser() -> ArgumentParser:
    def regex(pattern_str: str) -> re.Pattern:
        try:
            return re.compile(pattern_str)
        except re.error as e:
            raise ValueError(f"Invalid REGEX pattern: {e}")

    def regex_colon_path(regex_path: str) -> tuple[re.Pattern, str]:
        pattern_str, path = regex_path.rsplit(":", maxsplit=1)
        if any(not part.isidentifier() for part in path.split(".")):
            raise ValueError(f"Invalid PATH: {path}")
        return regex(pattern_str), path

    parser = ArgumentParser(
        prog="pybind11-stubgen", description="Generates stubs for specified modules"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        help="The root directory for output stubs",
        default="./stubs",
    )
    parser.add_argument(
        "--root-suffix",
        type=str,
        default=None,
        dest="root_suffix",
        help="Top-level module directory suffix",
    )

    parser.add_argument(
        "--ignore-invalid-expressions",
        metavar="REGEX",
        default=None,
        type=regex,
        help="Ignore invalid expressions matching REGEX",
    )
    parser.add_argument(
        "--ignore-invalid-identifiers",
        metavar="REGEX",
        default=None,
        type=regex,
        help="Ignore invalid identifiers matching REGEX",
    )

    parser.add_argument(
        "--ignore-unresolved-names",
        metavar="REGEX",
        default=None,
        type=regex,
        help="Ignore unresolved names matching REGEX",
    )

    parser.add_argument(
        "--ignore-all-errors",
        default=False,
        action="store_true",
        help="Ignore all errors during module parsing",
    )

    parser.add_argument(
        "--enum-class-locations",
        dest="enum_class_locations",
        metavar="REGEX:LOC",
        action="append",
        default=[],
        type=regex_colon_path,
        help="Locations of enum classes in "
        "<enum-class-name-regex>:<path-to-class> format. "
        "Example: `MyEnum:foo.bar.Baz`",
    )

    numpy_array_fix = parser.add_mutually_exclusive_group()
    numpy_array_fix.add_argument(
        "--numpy-array-wrap-with-annotated",
        default=False,
        action="store_true",
        help="Replace numpy/scipy arrays of "
        "'ARRAY_T[TYPE, [*DIMS], *FLAGS]' format with "
        "'Annotated[ARRAY_T, TYPE, FixedSize|DynamicSize(*DIMS), *FLAGS]'",
    )

    numpy_array_fix.add_argument(
        "--numpy-array-remove-parameters",
        default=False,
        action="store_true",
        help="Replace 'numpy.ndarray[...]' with 'numpy.ndarray'",
    )

    parser.add_argument(
        "--print-invalid-expressions-as-is",
        default=False,
        action="store_true",
        help="Suppress the replacement with '...' of invalid expressions"
        "found in annotations",
    )

    parser.add_argument(
        "--print-safe-value-reprs",
        metavar="REGEX",
        default=None,
        type=regex,
        help="Override the print-safe check for values matching REGEX",
    )

    parser.add_argument(
        "--exit-code",
        action="store_true",
        dest="exit_code",
        help="On error exits with 1 and skips stub generation",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        help="Don't write stubs. Parses module and report errors",
    )

    parser.add_argument(
        "--stub-extension",
        type=str,
        default="pyi",
        metavar="EXT",
        choices=["pyi", "py"],
        help="The file extension of the generated stubs. "
        "Must be 'pyi' (default) or 'py'",
    )

    parser.add_argument(
        "module_name",
        metavar="MODULE_NAME",
        type=str,
        help="module name",
    )

    return parser


def stub_parser_from_args(args) -> IParser:
    error_handlers_top: list[type] = [
        LoggerData,
        *([IgnoreAllErrors] if args.ignore_all_errors else []),
        *([IgnoreInvalidIdentifierErrors] if args.ignore_invalid_identifiers else []),
        *([IgnoreInvalidExpressionErrors] if args.ignore_invalid_expressions else []),
        *([IgnoreUnresolvedNameErrors] if args.ignore_unresolved_names else []),
    ]
    error_handlers_bottom: list[type] = [
        LogErrors,
        SuggestCxxSignatureFix,
        *([TerminateOnFatalErrors] if args.exit_code else []),
    ]

    numpy_fixes: list[type] = [
        *([FixNumpyArrayDimAnnotation] if args.numpy_array_wrap_with_annotated else []),
        *(
            [FixNumpyArrayRemoveParameters]
            if args.numpy_array_remove_parameters
            else []
        ),
    ]

    class Parser(
        *error_handlers_top,  # type: ignore[misc]
        FixMissing__future__AnnotationsImport,
        FixMissing__all__Attribute,
        FixMissingNoneHashFieldAnnotation,
        FixMissingImports,
        FilterTypingModuleAttributes,
        FixPEP585CollectionNames,
        FixTypingTypeNames,
        FixMissingFixedSizeImport,
        FixMissingEnumMembersAnnotation,
        OverridePrintSafeValues,
        *numpy_fixes,  # type: ignore[misc]
        FixNumpyDtype,
        FixNumpyArrayFlags,
        FixCurrentModulePrefixInTypeNames,
        FixBuiltinTypes,
        RewritePybind11EnumValueRepr,
        FilterClassMembers,
        ReplaceReadWritePropertyWithField,
        FilterInvalidIdentifiers,
        FixValueReprRandomAddress,
        FixRedundantBuiltinsAnnotation,
        FilterPybindInternals,
        FixRedundantMethodsFromBuiltinObject,
        RemoveSelfAnnotation,
        FixPybind11EnumStrDoc,
        ExtractSignaturesFromPybind11Docstrings,
        ParserDispatchMixin,
        BaseParser,
        *error_handlers_bottom,  # type: ignore[misc]
    ):
        pass

    parser = Parser()

    if args.enum_class_locations:
        parser.set_pybind11_enum_locations(dict(args.enum_class_locations))
    if args.ignore_invalid_identifiers is not None:
        parser.set_ignored_invalid_identifiers(args.ignore_invalid_identifiers)
    if args.ignore_invalid_expressions is not None:
        parser.set_ignored_invalid_expressions(args.ignore_invalid_expressions)
    if args.ignore_unresolved_names is not None:
        parser.set_ignored_unresolved_names(args.ignore_unresolved_names)
    if args.print_safe_value_reprs is not None:
        parser.set_print_safe_value_pattern(args.print_safe_value_reprs)
    return parser


def main():
    logging.basicConfig(
        level=logging.INFO,
        format="%(name)s - [%(levelname)7s] %(message)s",
    )
    args = arg_parser().parse_args()

    parser = stub_parser_from_args(args)
    printer = Printer(invalid_expr_as_ellipses=not args.print_invalid_expressions_as_is)

    out_dir, sub_dir = to_output_and_subdir(
        output_dir=args.output_dir,
        module_name=args.module_name,
        root_suffix=args.root_suffix,
    )

    run(
        parser,
        printer,
        args.module_name,
        out_dir,
        sub_dir=sub_dir,
        dry_run=args.dry_run,
        writer=Writer(stub_ext=args.stub_extension),
    )


def to_output_and_subdir(
    output_dir: Path, module_name: str, root_suffix: str | None
) -> tuple[Path, Path | None]:
    out_dir = Path(output_dir)

    module_path = module_name.split(".")

    if root_suffix is None:
        return out_dir.joinpath(*module_path[:-1]), None
    else:
        module_path = [f"{module_path[0]}{root_suffix}", *module_path[1:]]
        if len(module_path) == 1:
            sub_dir = Path(module_path[-1])
        else:
            sub_dir = None
        return out_dir.joinpath(*module_path[:-1]), sub_dir


def run(
    parser: IParser,
    printer: Printer,
    module_name: str,
    out_dir: Path,
    sub_dir: Path | None,
    dry_run: bool,
    writer: Writer,
):
    module = parser.handle_module(
        QualifiedName.from_str(module_name), importlib.import_module(module_name)
    )
    parser.finalize()

    if module is None:
        raise RuntimeError(f"Can't parse {module_name}")

    if dry_run:
        return

    out_dir.mkdir(exist_ok=True, parents=True)
    writer.write_module(module, printer, to=out_dir, sub_dir=sub_dir)


if __name__ == "__main__":
    main()



from __future__ import annotations

import abc
import types
from typing import Any

from pybind11_stubgen.parser.errors import ParserError
from pybind11_stubgen.structs import (
    Alias,
    Argument,
    Attribute,
    Class,
    Docstring,
    Field,
    Function,
    Import,
    InvalidExpression,
    Method,
    Module,
    Property,
    QualifiedName,
    ResolvedType,
    Value,
)


class IParser(abc.ABC):
    @abc.abstractmethod
    def handle_alias(self, path: QualifiedName, origin: Any) -> Alias | None:
        ...

    @abc.abstractmethod
    def handle_attribute(self, path: QualifiedName, attr: Any) -> Attribute | None:
        ...

    @abc.abstractmethod
    def handle_bases(
        self, path: QualifiedName, bases: tuple[type, ...]
    ) -> list[QualifiedName]:
        ...

    @abc.abstractmethod
    def handle_class(self, path: QualifiedName, class_: type) -> Class | None:
        ...

    @abc.abstractmethod
    def handle_class_member(
        self, path: QualifiedName, class_: type, obj: Any
    ) -> Docstring | Alias | Class | list[Method] | Field | Property | None:
        ...

    @abc.abstractmethod
    def handle_docstring(self, path: QualifiedName, doc: Any) -> Docstring | None:
        ...

    @abc.abstractmethod
    def handle_field(self, path: QualifiedName, field: Any) -> Field | None:
        ...

    @abc.abstractmethod
    def handle_function(self, path: QualifiedName, func: Any) -> list[Function]:
        ...

    @abc.abstractmethod
    def handle_import(self, path: QualifiedName, origin: Any) -> Import | None:
        ...

    @abc.abstractmethod
    def handle_method(self, path: QualifiedName, method: Any) -> list[Method]:
        ...

    @abc.abstractmethod
    def handle_module(
        self, path: QualifiedName, module: types.ModuleType
    ) -> Module | None:
        ...

    @abc.abstractmethod
    def handle_module_member(
        self, path: QualifiedName, module: types.ModuleType, obj: Any
    ) -> (
        Docstring | Import | Alias | Class | list[Function] | Attribute | Module | None
    ):
        ...

    @abc.abstractmethod
    def handle_property(self, path: QualifiedName, prop: Any) -> Property | None:
        ...

    @abc.abstractmethod
    def handle_type(self, type_: type) -> QualifiedName:
        ...

    @abc.abstractmethod
    def handle_value(self, value: Any) -> Value:
        ...

    @abc.abstractmethod
    def parse_args_str(self, args_str: str) -> list[Argument]:
        ...

    @abc.abstractmethod
    def parse_annotation_str(
        self, annotation_str: str
    ) -> ResolvedType | InvalidExpression | Value:
        ...

    @abc.abstractmethod
    def parse_value_str(self, value: str) -> Value | InvalidExpression:
        ...

    @abc.abstractmethod
    def report_error(self, error: ParserError) -> None:
        ...

    @abc.abstractmethod
    def finalize(self):
        ...



from __future__ import annotations

from pybind11_stubgen.structs import Identifier, QualifiedName


class ParserError(Exception):
    pass


class InvalidIdentifierError(ParserError):
    def __init__(self, name: Identifier, found_at: QualifiedName):
        super().__init__()
        self.name = name
        self.at = found_at

    def __str__(self):
        return f"Invalid identifier '{self.name}' at '{self.at}'"


class InvalidExpressionError(ParserError):
    def __init__(self, expression: str):
        super().__init__()
        self.expression: str = expression

    def __str__(self):
        return f"Invalid expression '{self.expression}'"


class NameResolutionError(ParserError):
    def __init__(self, name: QualifiedName):
        super().__init__()
        self.name = name

    def __str__(self):
        return f"Can't find/import '{self.name}'"



from __future__ import annotations

import sys
from dataclasses import dataclass
from dataclasses import field as field_
from typing import Tuple, Union

if sys.version_info[:2] >= (3, 8):
    from typing import Literal

    Modifier = Literal["static", "class", None]
else:
    from typing import Optional

    Modifier = Optional[str]


class Identifier(str):
    pass


class Decorator(str):
    pass


class Docstring(str):
    pass


@dataclass
class InvalidExpression:
    text: str

    def __str__(self):
        return f"Invalid python expression `{self.text}`"


class QualifiedName(Tuple[Identifier, ...]):
    """Fully Qualified Name"""

    @classmethod
    def from_str(cls, name: str) -> QualifiedName:
        return QualifiedName(Identifier(part) for part in name.split("."))

    def __str__(self):
        return ".".join(self)

    @property
    def parent(self) -> QualifiedName:
        return QualifiedName(self[:-1])


@dataclass
class Value:
    repr: str
    is_print_safe: bool = False  # `self.repr` is valid python and safe to print as is

    def __str__(self):
        return self.repr


@dataclass
class ResolvedType:
    name: QualifiedName
    parameters: list[ResolvedType | Value | InvalidExpression] | None = field_(
        default=None
    )

    def __str__(self):
        if self.parameters:
            param_str = "[" + ", ".join(str(p) for p in self.parameters) + "]"
        else:
            param_str = ""
        return f"{self.name}{param_str}"


@dataclass
class Alias:
    name: Identifier
    origin: QualifiedName


Annotation = Union[ResolvedType, Value, InvalidExpression]


@dataclass
class Attribute:
    name: Identifier
    value: Value | None
    annotation: Annotation | None = field_(default=None)


@dataclass
class Argument:
    name: Identifier | None
    pos_only: bool = field_(default=False)
    kw_only: bool = field_(default=False)
    variadic: bool = field_(default=False)  # *args
    kw_variadic: bool = field_(default=False)  # **kwargs
    default: Value | InvalidExpression | None = field_(default=None)
    annotation: Annotation | None = field_(default=None)

    def __str__(self):
        result = []
        if self.variadic:
            result += ["*"]
        if self.kw_variadic:
            result += ["**"]
        result += [f"{self.name}"]
        if self.annotation:
            result += [f": {self.annotation}"]
        if self.default:
            result += [f" = {self.default}"]

        return "".join(result)


@dataclass
class Function:
    name: Identifier
    args: list[Argument] = field_(default_factory=list)
    returns: Annotation | None = field_(default=None)
    doc: Docstring | None = field_(default=None)
    decorators: list[Decorator] = field_(default_factory=list)

    def __str__(self):
        return (
            f"{self.name}({', '.join(str(arg) for arg in self.args)}) -> {self.returns}"
        )


@dataclass
class Property:
    name: Identifier
    modifier: Modifier
    doc: Docstring | None = field_(default=None)
    getter: Function | None = field_(default=None)
    setter: Function | None = field_(default=None)


@dataclass
class Method:
    function: Function
    modifier: Modifier


@dataclass
class Field:
    attribute: Attribute
    modifier: Modifier


@dataclass
class Class:
    name: Identifier
    doc: Docstring | None = field_(default=None)
    bases: list[QualifiedName] = field_(default_factory=list)
    classes: list[Class] = field_(default_factory=list)
    fields: list[Field] = field_(default_factory=list)
    methods: list[Method] = field_(default_factory=list)
    properties: list[Property] = field_(default_factory=list)
    aliases: list[Alias] = field_(default_factory=list)


@dataclass(eq=True, frozen=True)
class Import:
    name: Identifier | None
    origin: QualifiedName


@dataclass
class Module:
    name: Identifier
    doc: Docstring | None = field_(default=None)
    classes: list[Class] = field_(default_factory=list)
    functions: list[Function] = field_(default_factory=list)
    sub_modules: list[Module] = field_(default_factory=list)
    attributes: list[Attribute] = field_(default_factory=list)
    imports: set[Import] = field_(default_factory=set)
    aliases: list[Alias] = field_(default_factory=list)
